import scala.collection.immutable.Set
import scala.collection.immutable.List

object QBF {

  abstract class Formula
  case class And(lhs: Formula, rhs: Formula) extends Formula {
    override def toString = "(" + lhs.toString + " & " + rhs.toString + ")"
  }
  case class Not(f:Formula) extends Formula {
    override def toString = "!" + f.toString
  }
  case class Or(lhs: Formula, rhs: Formula) extends Formula {
    override def toString = "(" + lhs.toString + " | " + rhs.toString + ")"
  }
  case class Implies(lhs: Formula, rhs: Formula) extends Formula {
    override def toString = "(" + lhs.toString + " --> " + rhs.toString + ")"
  }
  case class Variable(name:String) extends Formula {
    override def toString = name
  }
  case object False extends Formula  {
    override def toString = "false"
  }
  case object True extends Formula {
    override def toString = "true"
  }
  case class Exists(l:Variable, f: Formula) extends Formula {
    override def toString = "(EX " + l.toString + " . " + f.toString + ")"  
  }
  case class Forall(l:Variable, f: Formula) extends Formula {
    override def toString = "(ALL " + l.toString + " . " + f.toString + ")"  
  }

  //Applies as much transformations to truth values as possible
  //For example
  //(true | x) ==> true
  //(y & false) ==> false
  //(False --> y) ==> true
  //(z | (False --> w) ==> true
  //etc...
  def simplify(f: Formula) : Formula = {
    f match {
      case And(lhs, rhs) => ...
      case Not(f1) => ...
      case Or(lhs, rhs) => ...
    }
  }

  //substitutes a formula for a literal in a quantifier-free formula
  //For example
  //subst(x, x, True) ==> True
  //subst((true | x), x, True) ==> true | true
  //subst(y, x, False) ==> y
  //subst(EX y . !y, y, False) is undefined
  //etc...
  def subst(f: Formula, l: Variable, replacement: Formula) : Formula = {
    f match ...
  }

  //takes a formula f and returns an equivalent formula that doesn't contain any quantifier
  //For example
  //elimQuantifiers(EX x . x) ==> true | false
  //elimQuantifiers(ALL x . x) ==> true & false
  //elimQuantifiers(ALL x . x --> y) ==> (true --> y) & (false --> y)
  //elimQuantifiers(EX y . !y) ==> !true | !false
  //etc...
  def elimQuantifiers(f: Formula) : Formula = {
    f match ...
  }

  //freeVars returns the set of free variables contained in its first argument
  //For example
  //freeVars(EX x . x) ==> {}
  //freeVars(x) ==> {x}
  //freeVars(EX w . (z | (False --> w))) ==> {z}
  def freeVars(f: Formula, boundVars: Set[Variable]) : Set[Variable] = {
    f match ...
  }

  //with the right q, "close" may universally or existentially close f.
  def close(f: Formula, q: (Variable, Formula) => Formula) : Formula = {
    ...
  }

  //For example
  //univClose(x) ==> ALL x . x
  //univClose(y --> x) ==> ALL x . ALL y . y --> x
  //univClose(EX w . x | w) ==> ALL x . EX w . x | w
  def univClose(f: Formula) : Formula = close(...)

  //For example
  //existClose(x) ==> EX x . x
  //existClose(y --> x) ==> EX x . EX y . (y --> x)
  //existClose(EX w . x | w) ==> EX x . EX w . (x | w)
  def existClose(f: Formula) : Formula = close(...)

  def test(f: Formula, yes: String, no: String, cf: Formula => Formula) : String = elimQuantifiers(cf(f)) match {
    case True => yes
    case False => no
    case _ => throw new RuntimeException("quantifier elimination did not return a truth value")
  }

  def isValid(f: Formula) : String = test(f, "VALID", "INVALID", (univClose(_)))

  def isSat(f: Formula) : String = test(f, "SAT", "UNSAT", (existClose(_)))

  def equiv(f1: Formula, f2: Formula) : String = isValid(And(Implies(f1, f2), Implies(f2, f1)))

  def main(args: Array[String]) {
    val testInputs = List(Forall(Variable("x"), True), 
      Exists(Variable("x"),Variable("x")),
      Exists(Variable("x"),Not(Variable("x"))),
      Variable("x"),
      Not(Variable("x")),
      Forall(Variable("x"),Variable("x")),
      Implies(Variable("x"),Variable("y")),
      Implies(False,Variable("y")),
      Exists(Variable("y"), And(Not(Or(Variable("x"), Variable("y"))), Implies(Variable("y"), Variable("x")))),
      Forall(Variable("x"), Not(Exists(Variable("y"), And(Not(Or(Variable("x"), Variable("y"))), Implies(Variable("y"), Variable("x"))))))) 

    testInputs foreach ((f:Formula) => println(f + ": " + isSat(f) + " and " + isValid(f)))
  }

  /*
  SHOULD DISPLAY:
  (ALL x . true): SAT and VALID
  (EX x . x): SAT and VALID
  (EX x . !x): SAT and VALID
  x: SAT and INVALID
  !x: SAT and INVALID
  (ALL x . x): UNSAT and INVALID
  (x --> y): SAT and INVALID
  (false --> y): SAT and VALID
  (EX y . (!(x | y) & (y --> x))): SAT and INVALID
  (ALL x . !(EX y . (!(x | y) & (y --> x)))): UNSAT and INVALID
  */

}
